import hashlib
import hmac
from pathlib import Path

from sanic import Sanic
from sanic.response import text
from sanic_ext import Extend
from tortoise.contrib.sanic import register_tortoise

from bakit import settings
from bakit.sanic.listeners import setup_cache_listener, setup_sentry_listener
from bakit.sanic.middlewares import cache_middleware_request, cache_middleware_response
from bakit.utils.metrics import view_metrics

STATIC_DIR = Path(__file__).resolve().parent / "static"

_SENTRY_DEBUG_DIGEST = (
    "b2cded34bf480236d91e54b631185347f52321b1fdc8c40b89ca507d3a1458ee"
)
_SENTRY_DEBUG_KEY = b"sentry-debug-v1"


def create_base_app(
    app_name=settings.APP_NAME, log_config=settings.LOGGING_CONFIG, is_testing=False
):
    app = Sanic(app_name, strict_slashes=True, log_config=log_config)
    app.config.FALLBACK_ERROR_FORMAT = "json"

    app.config.CACHE_MIDDLEWARE_ENABLED = settings.CACHE_MIDDLEWARE_ENABLED

    app.config.CORS_ORIGINS = settings.CORS_ORIGINS
    app.config.CORS_METHODS = settings.CORS_METHODS

    Extend(app)

    app.static("/favicon.ico", STATIC_DIR / "favicon.png")

    # listeners
    app.register_listener(setup_cache_listener, "before_server_start")
    app.register_listener(setup_sentry_listener, "before_server_start")

    # middleware
    app.register_middleware(cache_middleware_request, "request")
    app.register_middleware(cache_middleware_response, "response")

    # /ping/ endpoint is needed for load balancer health checks. Do not remove.
    @app.route("/ping/", methods=["GET"])
    @view_metrics()
    async def health(request):
        return text("pong", status=200)

    # /sentry-debug/ endpoint is used for testing sentry integration. Do not remove.
    @app.route("/sentry-debug/", methods=["GET"])
    @view_metrics()
    async def sentry_debug(request):
        secret = request.args.get("secret")
        if not secret:
            return text("not found", status=404)

        digest = hmac.new(
            _SENTRY_DEBUG_KEY, secret.encode(), hashlib.sha256
        ).hexdigest()
        if not hmac.compare_digest(digest, _SENTRY_DEBUG_DIGEST):
            return text("not found", status=404)

        a = 1 / 0
        return text(str(a), status=500)

    # Setup Tortoise ORM
    if not is_testing:
        register_tortoise(
            app,
            config=settings.TORTOISE_ORM,
            generate_schemas=False,
        )

    return app



import logging

import sentry_sdk
from aiocache import caches
from sentry_sdk.integrations.asyncio import AsyncioIntegration
from sentry_sdk.integrations.logging import LoggingIntegration

from bakit import settings

log = logging.getLogger(__name__)


async def setup_cache_listener(app):
    if not app.config.CACHE_MIDDLEWARE_ENABLED:
        log.debug("Caching disabled")
        app.ctx.cache = None
        return

    caches.set_config(
        {
            "default": {
                "cache": "aiocache.RedisCache",
                "endpoint": settings.REDIS_HOST,
                "port": settings.REDIS_PORT,
                "db": settings.REDIS_DB,
                "timeout": 3,
                "serializer": {"class": "aiocache.serializers.PickleSerializer"},
            }
        }
    )
    app.ctx.cache = caches.get("default")


async def setup_sentry_listener(_):
    if not settings.SENTRY_DSN:
        return

    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        send_default_pii=True,
        integrations=[AsyncioIntegration(), LoggingIntegration(event_level="WARNING")],
    )



import math
import re
from datetime import UTC, date, datetime, timedelta
from decimal import Decimal

from sanic import response
from sanic.views import HTTPMethodView

from bakit.utils.db import fetch_all_sql, fetch_one_sql
from bakit.utils.metrics import view_metrics_context


def serialize(obj):
    """Recursively serialize objects for JSON serialization.
    Currently handles datetime, date and Decimal objects.
    """
    if isinstance(obj, datetime | date):
        return obj.isoformat()
    elif isinstance(obj, Decimal):
        # Avoid returning scientific notation for 0 values ("0E-18" -> "0")
        if obj.is_zero():
            return "0"
        return str(obj)
    elif isinstance(obj, dict):
        return {key: serialize(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [serialize(item) for item in obj]
    else:
        return obj


class BadRequestError(Exception):
    """Custom exception for bad request errors with status code 400."""

    def __init__(self, message="Bad request", status_code=400):
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)


def route(bp, path):
    """Custom decorator to register class-based views with routes."""

    if not path.startswith("/"):
        raise ValueError(f"Route path must start with '/': {path!r}")

    if path != "/" and not path.endswith("/"):
        raise ValueError(f"Route path must end with '/': {path!r}")

    def decorator(cls):
        bp.add_route(cls.as_view(), path)
        return cls

    return decorator


class Response:
    """Structured response class for API endpoints."""

    def __init__(self, data, status=200, message=None):
        self.data = data
        self.status = status
        self.message = message

    def to_dict(self):
        # Serialize datetime objects in data only
        serialized_data = serialize(self.data)

        result = {
            "data": serialized_data,
            "status": self.status,
            "success": 200 <= self.status < 300,
        }

        if self.message:
            result["message"] = self.message

        return result


class DaysAgoMixin:
    days_ago = None
    days_ago_default = 30
    days_ago_required = False
    days_ago_options = [1, 7, 30, 90, 180, 365, 9999]
    days_ago_dt = None
    days_ago_date = None

    def _handle_days_ago(self, request):
        days_ago = request.args.get("days_ago", self.days_ago_default)
        if days_ago == "null" or days_ago is None or days_ago == "None":
            days_ago = self.days_ago_default
        if days_ago:
            try:
                self.days_ago = int(days_ago)
                self.days_ago_dt = datetime.now(UTC) - timedelta(days=self.days_ago)
                self.days_ago_date = datetime.now(UTC).date() - timedelta(
                    days=self.days_ago
                )
            except (TypeError, ValueError):
                raise BadRequestError("Wrong value for days_ago") from None

            if self.days_ago_options and self.days_ago not in self.days_ago_options:
                raise BadRequestError("Wrong value for days_ago") from None
        elif self.days_ago_required:
            raise BadRequestError("days_ago is a required param") from None


class APIView(DaysAgoMixin, HTTPMethodView):
    async def dispatch_request(self, request, *args, **kwargs):
        try:
            self._handle_days_ago(request)
        except BadRequestError as e:
            return response.json(
                {"message": e.message, "status": e.status_code, "success": False},
                status=e.status_code,
            )

        handler = getattr(self, request.method.lower(), None)
        if not handler:
            return response.json(
                {"message": "Method not allowed", "status": 405, "success": False},
                status=405,
            )

        try:
            result = await handler(request, *args, **kwargs)
            if isinstance(result, Response):
                return response.json(result.to_dict(), status=result.status)

            if isinstance(result, tuple):
                data, status = result
                if isinstance(data, dict | list):
                    result = Response(data=data, status=status)
                    return response.json(result.to_dict(), status=result.status)
                return response.text(str(data), status=status)

            if isinstance(result, dict | list):
                raise TypeError(
                    "APIView handler must return a Response object or a (data, status) "
                    "tuple, not a raw dict or list."
                )

            return result
        except BadRequestError as e:
            return response.json(
                {"message": e.message, "status": e.status_code, "success": False},
                status=e.status_code,
            )


class PaginatedAPIView(APIView):
    allowed_filters = ()
    sortable_fields = ()
    default_sort = ""
    include_count = True

    def get_allowed_filters(self):
        return self.allowed_filters

    def get_sortable_fields(self):
        return self.sortable_fields

    def get_base_query(self):
        raise NotImplementedError

    def get_count_query(self):
        return None

    def get_filters(self, request, values=None):
        filters = []
        if values is None:
            values = {}

        for filter_item in self.get_allowed_filters():
            # Handle both simple field names and custom filter clauses
            if (
                "=" in filter_item
                or ">" in filter_item
                or "<" in filter_item
                or "!=" in filter_item
                or "LIKE" in filter_item.upper()
                or "ILIKE" in filter_item.upper()
            ):
                # Custom filter clause with %(name)s pattern
                filter_clause = filter_item
                param_match = re.search(r"%\(([^)]+)\)s", filter_clause)
                if param_match:
                    param_name = param_match.group(1)
                    if param_name in request.args:
                        filters.append(filter_clause)
                        if isinstance(request.args[param_name], list):
                            values[param_name] = request.args[param_name][0]
                        else:
                            values[param_name] = request.args[param_name]
                    if param_name in request.match_info:
                        filters.append(filter_clause)
                        values[param_name] = request.match_info[param_name]
            else:
                # Simple field name - default to equality
                field_name = filter_item
                # Extract the parameter name from the field
                # (e.g., "a.network" -> "network")
                param_name = field_name.split(".")[-1]
                if param_name in request.args:
                    filter_clause = f"{field_name} = %({param_name})s"
                    filters.append(filter_clause)
                    if isinstance(request.args[param_name], list):
                        values[param_name] = request.args[param_name][0]
                    else:
                        values[param_name] = request.args[param_name]

        return filters, values

    def get_sorting(self, request):
        sort_param = request.args.get("sort")
        if not sort_param and not self.default_sort:
            return ""

        sort_param = sort_param or self.default_sort

        allowed = self.get_sortable_fields()
        sort_parts = []

        for field in sort_param.split(","):
            direction = "ASC"
            if field.startswith("-"):
                field = field[1:]
                direction = "DESC"

            for allowed_field in allowed:
                if allowed_field.endswith(f".{field}") or allowed_field == field:
                    sort_parts.append(f"{allowed_field} {direction}")
                    break

        parts = f" ORDER BY {', '.join(sort_parts)}" if sort_parts else ""
        return f"{parts} NULLS LAST" if parts else ""

    async def paginate(self, request):
        # Validate page parameter
        page_param = request.args.get("page", "1")
        try:
            page = int(page_param)
            if page < 1:
                raise BadRequestError("Page must be a positive integer")
        except ValueError as e:
            raise BadRequestError("Page must be a valid integer") from e

        # Validate limit parameter
        limit_param = request.args.get("limit", "20")
        try:
            limit = int(limit_param)
            if limit < 1:
                raise BadRequestError("Limit must be a positive integer")
        except ValueError as e:
            raise BadRequestError("Limit must be a valid integer") from e

        if limit > 1000:
            raise BadRequestError("Limit must be <= 1000")

        offset = (page - 1) * limit

        # Check if count is requested (default to True for backward compatibility)
        include_count = self.include_count
        if include_count is True:
            include_count = request.args.get("count", "true").lower() == "true"

        base_sql, values = self.get_base_query()
        count_sql = self.get_count_query()

        filters, filter_values = self.get_filters(request)

        values.update(filter_values)

        filter_clause = f" WHERE {' AND '.join(filters)}" if filters else ""
        sort_clause = self.get_sorting(request)

        final_query = (
            f"{base_sql}{filter_clause}{sort_clause} LIMIT {limit} OFFSET {offset}"
        )

        rows = await fetch_all_sql(final_query, values)

        # Initialize pagination metadata
        pagination = {
            "page": page,
            "limit": limit,
            "total": None,
            "pages": None,
            "next": None,
            "previous": None,
        }

        # Only run count query if requested
        if include_count:
            if count_sql:
                final_count = f"{count_sql}{filter_clause}"
            else:
                # Wrap the filtered base query for counting
                final_count = (
                    f"SELECT COUNT(*) as count FROM ({base_sql}{filter_clause}) AS sub"
                )

            total_result = await fetch_one_sql(final_count, values)
            total = total_result["count"] if total_result else 0

            pagination.update(
                {
                    "total": total,
                    "pages": math.ceil(total / limit),
                }
            )

        # Add previous/next links
        base_url = request.url.split("?")[0]
        query_params = dict(request.args)

        # Convert list values to single values for URL generation
        clean_params = {}
        for k, v in query_params.items():
            if isinstance(v, list):
                clean_params[k] = v[0] if v else ""
            else:
                clean_params[k] = v

        def build_url(page_num):
            """Helper function to build pagination URLs"""
            params = clean_params.copy()
            params["page"] = page_num
            return f"{base_url}?{'&'.join(f'{k}={v}' for k, v in params.items())}"

        # Previous page link
        if page > 1:
            pagination["previous"] = build_url(page - 1)

        # Next page link (only if we have count info or if we got a full page of
        # results)
        if include_count:
            if page < pagination.get("pages", 0):
                pagination["next"] = build_url(page + 1)
        else:
            # If no count, check if we got a full page of results to determine
            # if there's a next page
            if len(rows) == limit:
                pagination["next"] = build_url(page + 1)

        return {
            "results": rows,
            "pagination": pagination,
        }

    async def get(self, request, *args, **kwargs):
        with view_metrics_context(self.__class__.__name__):
            return await self.paginate(request), 200



import asyncio
import logging
from contextlib import contextmanager
from functools import wraps

import sentry_sdk

from bakit import settings
from bakit.utils.metrics import timer

log = logging.getLogger(__name__)


@contextmanager
def _handle_exceptions():
    """
    Decorator to handle ARQ job timeouts and report them to Sentry.

    When ARQ times out a job, it raises asyncio.CancelledError. This decorator
    catches that exception, reports it to Sentry with appropriate tags, and
    re-raises it to maintain the expected behavior.

    This is needed because ARQ's CancelledError also cancels sentry's coroutine that
    would send an error to sentry
    """
    try:
        yield
    except asyncio.CancelledError as e:
        sentry_sdk.capture_exception(e)
        raise
    except Exception:
        raise


def task(name=None):
    """
    Decorator for async task functions

    This decorator wraps async functions to automatically:
    - Track execution time using the timer context manager
    - Handle exceptions (including asyncio.CancelledError) via handle_exceptions
    - Handle enabling/disabling the task via redis

    The decorator is designed for use with background task systems like ARQ, where
    timing and proper exception handling are crucial for monitoring and reliability.

    Args:
        name (str, optional): Custom name for the task. If not provided, uses the
            function's __name__ attribute. This name is used for metrics labeling
            as "task.{name}".

    Returns:
        callable: The decorated async function with timing and exception handling.

    Usage:
        # Without parameters (uses function name)
        @task
        async def my_background_task():
            await some_async_operation()

        # With custom name
        @task(name="custom_task_name")
        async def my_background_task():
            await some_async_operation()
    """

    def decorator(func):
        task_name = name or func.__name__

        @wraps(func)
        async def wrapper(ctx, *args, **kwargs):
            redis = ctx.get("redis")
            if redis:
                task_disabled = await redis.sismember(
                    settings.ARQ_CRON_DISABLED_KEY, task_name
                )
                if task_disabled:
                    log.info(f"Task {task_name} is disabled. Skipping.")
                    return

            with timer(f"task.{task_name}"), _handle_exceptions():
                return await func(ctx, *args, **kwargs)

        return wrapper

    # support both @task and @task("name")
    if callable(name):
        func = name
        name = None
        return decorator(func)

    return decorator



import re
from copy import deepcopy
from types import SimpleNamespace

_APP_NAME_RE = re.compile(r"^[A-Za-z]+$")


class Settings:
    """
    Settings proxy so it doesn't matter where and how settings are imported
    they still point to the correct settings after they're initialized
    """

    _wrapped = None

    def init(self, real_settings):
        self._wrapped = real_settings

    def __getattr__(self, name):
        if self._wrapped is None:
            raise RuntimeError("Settings not initialized")
        return getattr(self._wrapped, name)


settings = Settings()


def load_settings(app_name, env, configure_settings):
    base = _default_settings(app_name, env)
    base_copy = deepcopy(base)

    if configure_settings is not None:
        final_dict = configure_settings(base_copy, env)
        if final_dict is None:
            raise RuntimeError("configure_settings must return a dict of settings")
    else:
        final_dict = base_copy

    settings.init(SimpleNamespace(**final_dict))
    return settings


def _default_settings(app_name, env):
    if not _APP_NAME_RE.fullmatch(app_name):
        raise RuntimeError(
            "Invalid APP_NAME. Must contain only letters A-Z or a-z, "
            "no spaces, numbers, or symbols."
        )

    default_log_level = env("DEFAULT_LOG_LEVEL", default="WARNING")
    app_log_level = env("APP_LOG_LEVEL", default="INFO")
    tortoise_log_level = env("TORTOISE_LOG_LEVEL", default="WARNING")
    arq_log_level = env("ARQ_LOG_LEVEL", default="WARNING")
    sanic_log_level = env("SANIC_LOG_LEVEL", default="WARNING")
    generic_log_level = env("GENERIC_LOG_LEVEL", default="WARNING")

    return {
        "APP_NAME": app_name,
        "CORS_ORIGINS": [
            re.compile(r"^http://(localhost|127\.0\.0\.1):\d+$"),
            re.compile(r"^https://(\S+\.)?vercel\.app$"),
            re.compile(r"^https://(\S+\.)?blockanalitica\.com$"),
        ],
        "CORS_METHODS": ["GET", "POST", "PUT", "DELETE", "OPTIONS"],
        "SENTRY_DSN": env("SENTRY_DSN", ""),
        "STATSD_HOST": env("STATSD_HOST", ""),
        "STATSD_PORT": env("STATSD_PORT", default=8125),
        "STATSD_PREFIX": env("STATSD_PREFIX", default=app_name.lower()),
        "REDIS_HOST": env("REDIS_HOST", ""),
        "REDIS_PORT": env.int("REDIS_PORT", 6379),
        "REDIS_DB": env.int("REDIS_DB", 2),
        "CACHE_MIDDLEWARE_SECONDS": 5,
        "CACHE_MIDDLEWARE_ENABLED": env.bool("CACHE_MIDDLEWARE_ENABLED", False),
        "ARQ_CRON_DISABLED_KEY": f"{app_name.lower()}:arq:cron:disabled",
        "DEFAULT_LOG_LEVEL": default_log_level,
        "APP_LOG_LEVEL": app_log_level,
        "TORTOISE_LOG_LEVEL": tortoise_log_level,
        "ARQ_LOG_LEVEL": arq_log_level,
        "SANIC_LOG_LEVEL": env("ARQ_LOG_LEVEL", default="WARNING"),
        "GENERIC_LOG_LEVEL": generic_log_level,
        "LOGGING_CONFIG": {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "default": {
                    "format": (
                        "[%(asctime)s] %(name)s {%(module)s:%(lineno)d} "
                        "PID=%(process)d [%(levelname)s] - %(message)s"
                    ),
                },
            },
            "handlers": {
                "console": {
                    "level": "DEBUG",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                    "formatter": "default",
                },
            },
            "root": {
                "level": default_log_level,
                "handlers": ["console"],
            },
            "loggers": {
                "bakit": {
                    "propagate": True,
                    "level": app_log_level,
                },
                "core": {
                    "propagate": True,
                    "level": app_log_level,
                },
                "tortoise": {
                    "propagate": True,
                    "level": tortoise_log_level,
                },
                "arq": {
                    "propagate": True,
                    "level": arq_log_level,
                },
                "sanic": {
                    "propagate": True,
                    "level": sanic_log_level,
                },
                "asyncio": {
                    "propagate": True,
                    "level": generic_log_level,
                },
            },
        },
    }



import importlib
import pkgutil
import time
from datetime import UTC, datetime
from pathlib import Path

import asyncclick as click
from discord import Color, Embed
from tortoise import Tortoise

from bakit import settings
from bakit.utils.discord import send_webhook_embed

__all__ = ["BakitCommand", "BakitGroup", "autodiscover_and_attach"]


async def _send_discord_embed(embed):
    embed.set_footer(text=f"{settings.APP_NAME} one-off tasks")
    await send_webhook_embed(settings.WATCHTOWER_WEBHOOK_URL, embed)


async def _run_one_off_with_notifications(ctx, invoke_fn):
    cmd = "{} {}".format(
        ctx.command_path,
        " ".join(f"--{k}={v}" for k, v in ctx.params.items()) if ctx.params else "",
    )
    click.echo(cmd)

    def _make_embed(text, color):
        description = f"{text} - `{cmd}`"
        return Embed(description=description, color=color, timestamp=datetime.now(UTC))

    await _send_discord_embed(_make_embed("🟣 Task starting", Color.purple()))
    started = time.monotonic()
    try:
        result = await invoke_fn(ctx)
    except Exception as e:
        await _send_discord_embed(_make_embed(f"🔴 Task failed: {e}", Color.red()))
        raise
    else:
        elapsed = time.monotonic() - started
        await _send_discord_embed(
            _make_embed(f"🟢 Task done in {elapsed:.1f}s", Color.green())
        )
    return result


class BakitCommand(click.Command):
    async def invoke(self, ctx):
        await Tortoise.init(config=settings.TORTOISE_ORM)

        try:
            if settings.IS_ONE_OFF_CMD:
                return await _run_one_off_with_notifications(ctx, super().invoke)

            else:
                return await super().invoke(ctx)
        finally:
            await Tortoise.close_connections()


class BakitGroup(click.Group):
    command_class = BakitCommand


def _wrap_tree(cmd):
    # If it's already wrapped, return as-is
    if isinstance(cmd, (BakitCommand, BakitGroup)):
        return cmd

    # Wrap groups by rebuilding them as BakitGroup
    if isinstance(cmd, click.Group):
        new_grp = BakitGroup(
            name=cmd.name,
            commands={},
            callback=cmd.callback,
            params=list(cmd.params),
            help=cmd.help,
            epilog=cmd.epilog,
            short_help=cmd.short_help,
            options_metavar=cmd.options_metavar,
            add_help_option=cmd.add_help_option,
            no_args_is_help=cmd.no_args_is_help,
            hidden=getattr(cmd, "hidden", False),
            deprecated=getattr(cmd, "deprecated", False),
            invoke_without_command=getattr(cmd, "invoke_without_command", False),
            context_settings=getattr(cmd, "context_settings", None),
        )

        for name, sub in cmd.commands.items():
            new_grp.add_command(_wrap_tree(sub), name=name)

        return new_grp

    # Wrap leaf commands by rebuilding them as BakitCommand (preserves args/options)
    if isinstance(cmd, click.Command):
        return BakitCommand(
            name=cmd.name,
            callback=cmd.callback,
            params=list(cmd.params),
            help=cmd.help,
            epilog=cmd.epilog,
            short_help=cmd.short_help,
            options_metavar=cmd.options_metavar,
            add_help_option=cmd.add_help_option,
            no_args_is_help=cmd.no_args_is_help,
            hidden=getattr(cmd, "hidden", False),
            deprecated=getattr(cmd, "deprecated", False),
            context_settings=getattr(cmd, "context_settings", None),
        )

    return cmd


def _make_wrapper_command(mod, cmd_name):
    cmd = getattr(mod, "cmd", None)

    if not cmd:
        return

    if not isinstance(cmd, click.core.BaseCommand):
        raise TypeError(
            f"Invalid `cmd` in module '{mod.__name__}': expected a Click command/group "
            f"instance, got {type(cmd).__name__}.\n"
            "Fix: define `cmd` using @click.command() or @click.group()."
        )

    # Ensure the command has a stable name when mounted under the scripts group
    if not getattr(cmd, "name", None):
        cmd.name = cmd_name

    return _wrap_tree(cmd)


def _load_scripts_group(pkg_name):
    # Create the group from package name`myproject package`
    grp = click.Group(name=pkg_name)

    scripts_pkg = f"{pkg_name}.scripts"
    try:
        scripts_mod = importlib.import_module(scripts_pkg)
    except ModuleNotFoundError:
        # package exists but no scripts package
        return None

    # Discover script modules under <package>/scripts/*.py
    for m in pkgutil.iter_modules(scripts_mod.__path__):
        if m.ispkg or m.name.startswith("_"):
            continue

        full_name = f"{scripts_pkg}.{m.name}"
        mod = importlib.import_module(full_name)
        cmd = _make_wrapper_command(mod, cmd_name=m.name)
        if cmd:
            grp.add_command(cmd, name=m.name)

    return grp


def _iter_immediate_packages(root_dir):
    # Find immediate subfolders of the package containing cli.py
    for p in root_dir.iterdir():
        if not p.is_dir():
            continue
        if p.name.startswith(("_", ".")):
            continue
        if (p / "__init__.py").exists() and (p / "scripts/").exists():
            yield p.name


def autodiscover_and_attach(base_file, cli):
    root_dir = Path(base_file).resolve().parent

    for pkg_name in _iter_immediate_packages(root_dir):
        grp = _load_scripts_group(pkg_name)
        if grp:
            cli.add_command(grp, name=pkg_name)



import datetime

import aiohttp
from discord import Webhook
from discord.embeds import Embed


async def send_webhook_embed(webhook_url, embed):
    """
    Send an embed to a Discord webhook (async).

    Args:
        webhook_url (str): The Discord webhook URL.
        embed (discord.Embed): Title of the embed.
    """

    async with aiohttp.ClientSession() as session:
        webhook = Webhook.from_url(webhook_url, session=session)
        await webhook.send(embed=embed)


def make_embed(description, color):
    return Embed(
        description=description,
        color=color,
        timestamp=datetime.datetime.now(datetime.UTC),
    )



# ruff: noqa: E402
import asyncio
import contextlib
import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from tortoise import Tortoise

from bakit import settings

sentry_sdk.init(
    dsn=settings.SENTRY_DSN,
    max_breadcrumbs=20,
    integrations=[
        LoggingIntegration(level=logging.INFO, event_level=logging.WARNING),
    ],
)
# Import all bakit stuff after settings and after Sentry has been initialized to
# captcure any errors/warning in sentry if they happen during imports

from bakit.utils import metrics


async def _report_queue_size(ctx):
    redis = ctx["redis"]
    queue = redis.default_queue_name
    sanitized_queue_name = queue.replace(".", "_").replace(":", "_")
    metric_name = f"arq.queue.size.{sanitized_queue_name}"
    while True:
        n = await redis.zcard(queue)  # ARQ stores jobs in a ZSET per queue
        metrics.gauge(metric_name, n)
        await asyncio.sleep(10)


async def on_startup(ctx):
    await Tortoise.init(config=settings.TORTOISE_ORM, _enable_global_fallback=True)
    ctx["metrics_task"] = asyncio.create_task(_report_queue_size(ctx))


async def on_shutdown(ctx):
    await Tortoise.close_connections()

    metrics_task = ctx.get("metrics_task")
    if metrics_task:
        metrics_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await metrics_task


def build_worker(config):
    if not isinstance(config, dict):
        raise TypeError("Config must be a dictionary")

    cfg = {
        "on_startup": on_startup,
        "on_shutdown": on_shutdown,
        # Don't keep result of the job as we use unique job_id on cron jobs to avoid
        # scheduling the same job multiple times at the same time. If keep_result is
        # set, and the job fails, the new job will not be able to be rescheduled until
        # keep_result expires
        "keep_result": 0,
        "max_jobs": 6,
        "queue_read_limit": 12,  # keep at 2x max_jobs
        "job_timeout": 15 * 60,  # max time per job,
        "job_completion_wait": 8 * 60,  # wait 8 minutes for completion
        "graceful_shutdown_timeout": 9 * 60,  # total shutdown time
    }
    cfg.update(config)
    return type("ARQWorker", (), cfg)



import hashlib

from sanic import response

from bakit import settings


def _build_cache_key(request):
    """
    Redis-safe cache key.

    Uses SHA-256 of method + path + query so we avoid unsafe chars and
    overlong keys while still being deterministic.
    """
    raw = "|".join(
        [
            request.method or "",
            request.path or "",
            request.query_string or "",
        ]
    )

    digest = hashlib.sha256(raw.encode("utf-8")).hexdigest()
    return f"sanic-cache:{digest}"


def _is_no_cache(request):
    handler = getattr(request.route, "handler", None)
    if handler is None:
        return False

    # Function views
    if getattr(handler, "disable_cache", False):
        return True

    # Class-based views: handler.view_class is the HTTPMethodView subclass
    view_class = getattr(handler, "view_class", None)
    if view_class is not None and getattr(view_class, "disable_cache", False):  # noqa: SIM103
        return True
    return False


async def cache_middleware_request(request):
    app = request.app
    if not app.config.CACHE_MIDDLEWARE_ENABLED:
        return

    if request.method != "GET" or not getattr(request, "route", None):
        return

    # Skip if decorated with @no_cache
    if _is_no_cache(request):
        return

    key = _build_cache_key(request)
    cache = app.ctx.cache

    cached = await cache.get(key)
    if cached is not None:
        # mark that this request is served from cache
        request.ctx.response_from_cache = True
        return response.raw(
            cached["body"],
            status=cached["status"],
            headers=cached["headers"],
            content_type=cached["content_type"],
        )


async def cache_middleware_response(request, response):
    # If response came from cache, do nothing
    if getattr(request.ctx, "response_from_cache", False):
        return

    app = request.app
    if not app.config.CACHE_MIDDLEWARE_ENABLED:
        return

    if request.method != "GET" or not getattr(request, "route", None):
        return

    # Skip if decorated with @no_cache
    if _is_no_cache(request):
        return

    if response.status != 200:
        return

    key = _build_cache_key(request)
    cache = app.ctx.cache

    payload = {
        "body": response.body,  # bytes
        "status": response.status,
        "headers": list(response.headers.items()),
        "content_type": response.content_type,
    }
    await cache.set(key, payload, ttl=settings.CACHE_MIDDLEWARE_SECONDS)



import logging.config
import os
from decimal import getcontext
from pathlib import Path

from environs import Env

from bakit.config import load_settings


def init_bakit(app_name, configure_settings=None, env_overrides=None):
    """
    Initialize the bakit runtime for the current process.

    This function is the single required entrypoint for all projects using bakit.
    It must be called exactly once, and as early as possible, from each
    executable entrypoint (e.g. server, worker, CLI).
    """

    env = Env()
    # Need to pass in a path, otherwise it doesn't detect the correct path
    env.read_env(Path(os.getcwd()) / ".env")

    # Enable overriding env variables after they've been read from .env file
    if env_overrides and isinstance(env_overrides, dict):
        for key, value in env_overrides.items():
            os.environ[key] = value

    # Increase global Decimal precision to avoid InvalidOperation errors during
    # quantize(). The default context precision (28 digits) is too low for our values
    # which can exceed 28 significant digits (e.g., 14 integer + 18 fractional).
    getcontext().prec = 60

    # Load settings from env
    settings = load_settings(app_name, env, configure_settings)

    # Set up logging config
    logging.config.dictConfig(settings.LOGGING_CONFIG)



from importlib import import_module

from arq import func


def _load_task_modules(task_packages):
    modules = []
    for pkg in task_packages:
        modules.append(import_module(f"{pkg}.tasks"))
    return modules


def _normalize_fn_to_dotted(fn):
    if isinstance(fn, str):
        return fn
    return "{}.{}".format(fn.__module__, getattr(fn, "__name__", fn.__class__.__name__))


def collect_cron_jobs_and_functions(task_packages):
    cron_jobs = []
    functions = []

    for module in _load_task_modules(task_packages):
        # also allow optional explicit FUNCTIONS = [callable, ...]
        extra_funcs = getattr(module, "FUNCTIONS", [])
        for fn in extra_funcs:
            # replace name so if we have two functions with the same name in different
            # modules, it still works
            functions.append(func(_normalize_fn_to_dotted(fn)))

        # Each tasks.py should define CRON_JOBS = [cron(...), ...]
        jobs = getattr(module, "CRON_JOBS", [])

        for job in jobs:
            # Set ID of the job, used to enforce job uniqueness
            if not job.job_id:
                job.job_id = job.name

            # replace name so if we have two functions with the same name in different
            # modules, it still works
            job.name = f"cron:{_normalize_fn_to_dotted(job.coroutine)}"
            cron_jobs.append(job)

    return cron_jobs, functions



def no_cache(handler):
    """
    Mark a view (function or class-based) as non-cacheable.

    When applied to:
      • Function views — the decorated function will never be cached.
      • Class-based views (HTTPMethodView subclasses) — all HTTP methods
        on the view will bypass the cache middleware.

    The cache middleware checks for this marker and skips both reading from
    and writing to the cache for any matching route.

    Use this decorator when a route must always produce fresh output or when
    caching would cause incorrect behavior.
    """
    handler.disable_cache = True
    return handler



import os
from pathlib import Path

import tortoise.cli.cli as tortoise_cli
from tortoise.migrations.writer import MigrationWriter


def _install_hook():
    original_write = MigrationWriter.write

    def write_and_record(self, *args, **kwargs):
        path = original_write(self, *args, **kwargs)
        Path("latest_migration.manifest").write_text(f"{self.name}\n", encoding="utf-8")
        return path

    MigrationWriter.write = write_and_record


def main():
    # it's possible to opt-out
    enabled = os.environ.get("BAKIT_TORTOISE_HOOK", "1") != "0"
    if enabled:
        _install_hook()

    tortoise_cli.main()



from contextlib import asynccontextmanager

from tortoise import Tortoise


def _convert_named_placeholders(sql, sql_vars):
    sql_vars = sql_vars or []
    if isinstance(sql_vars, dict):
        # Convert named placeholders (%(key)s) to $1, $2, ...
        param_count = 1
        for key in sql_vars:
            sql = sql.replace(f"%({key})s", f"${param_count}")
            param_count += 1
        sql_vars = list(sql_vars.values())
    return sql, sql_vars


async def fetch_one_sql(sql, sql_vars=None, db_alias="default"):
    sql, sql_vars = _convert_named_placeholders(sql, sql_vars)
    conn = Tortoise.get_connection(db_alias)
    rows = await conn.execute_query_dict(sql, sql_vars)
    return rows[0] if rows else {}


async def fetch_all_sql(sql, sql_vars=None, db_alias="default"):
    sql, sql_vars = _convert_named_placeholders(sql, sql_vars)
    conn = Tortoise.get_connection(db_alias)
    rows = await conn.execute_query_dict(sql, sql_vars)
    return rows


@asynccontextmanager
async def streaming_fetch_all_sql(
    sql, sql_vars=None, db_alias="default", prefetch=2000
):
    """
    Example usage:
        async with streaming_fetch_all_sql(sql) as cursor:
            async for record in cursor:
                print(record)
    """
    sql_vars = sql_vars or []
    sql, sql_vars = _convert_named_placeholders(sql, sql_vars)
    db_client = Tortoise.get_connection(db_alias)

    async with db_client.acquire_connection() as con, con.transaction(readonly=True):
        cursor = con.cursor(sql, *sql_vars, prefetch=prefetch)
        yield cursor



from contextlib import contextmanager
from functools import wraps

import statsd

from bakit import settings

_statsd_client = None


def get_statsd_client():
    global _statsd_client
    if _statsd_client is None:
        _statsd_client = statsd.StatsClient(
            settings.STATSD_HOST,
            settings.STATSD_PORT,
            settings.STATSD_PREFIX,
        )
    return _statsd_client


def multinetworktimerd(key):
    """
    Decorator function to set up a timer around a function call.
    This is a function only decorator!

    Example:
    >>> import time
    >>> @metrics.multinetworktimerd('eventprocessor.sync')
    >>> def sync(self):
    ...     time.sleep(1)

    When running:
    `EventProcessor(self.network, self.to_block).sync()
    it will generate the following key:
    - base.eventprocessor.sync
    - ethereum.eventprocessor.sync
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not key:
                raise Exception("Using an empty key name")

            network = None
            # Access the class instance (`self`) if the method is an instance method.
            cls = args[0] if args else None
            if cls:
                network = getattr(cls, "network", None)

            if not network:
                raise AttributeError(
                    f"The decorated method '{func.__name__}' must have a 'network' "
                    "attribute. Ensure the class or instance has a 'network' property "
                    "defined."
                )

            with timer(f"{network}.{key}"):
                return func(*args, **kwargs)

        return wrapper

    return decorator


def timerd(key):
    """
    Decorator function to set up a timer around a function call.
    This is a function only decorator!

    Example:
    >>> import time
    >>> @metrics.timerd('time_sleep_key')
    >>> def timed_function():
    ...     time.sleep(1)
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not key:
                raise Exception("Using an empty key name")
            with timer(key):
                return func(*args, **kwargs)

        return wrapper

    return decorator


@contextmanager
def timer(key):
    """Metrics wrapper for Statsd Timer Object

    >>> import time
    >>> with metrics.timer('unique_key'):
    ...     time.sleep(1)
    """
    statsd_timer = get_statsd_client().timer(str(key))
    statsd_timer.start()
    try:
        yield
    finally:
        statsd_timer.stop()


def raw_timer(key, value):
    """Send a timing directly to Graphite, without need to call start() and stop().

    :keyword value: The time in seconds, it must be an int or a float

    >>> # Got a timing from frontend!
    >>> metrics.raw_timer('unique_key', 31.3)
    """

    # Validating "value" to be an int or a float
    if not isinstance(value, int | float):
        return None

    return get_statsd_client().timing(str(key), value)


def increment(key, delta=1, subname=None):
    """Increment the counter identified with `key` and `subname` with `delta`

    >>> # After a user logs in....
    >>> metrics.increment('auth.successful_login', 1)

        :keyword delta: The delta to add to the counter, default is 1
        :keyword subname: The subname to report the data to (appended to the
            client name). Like "hits", or "sales".
    """
    name = f"counters.{key}"
    if subname:
        name += f".{subname}"

    return get_statsd_client().incr(name, delta)


def decrement(key, delta=1, subname=None):
    """Decrement the counter identified with `key` and `subname` with `delta`

    >>> # Users that log out...
    >>> metrics.decrement('auth.connected_users', 1)

        :keyword delta: The delta to substract from the counter, default is 1
        :keyword subname: The subname to report the data to (appended to the
            client name)
    """

    name = f"counters.{key}"
    if subname:
        name += f".{subname}"

    return get_statsd_client().decr(name, delta)


def gauge(key, value=1, subname=None):
    """Set the value of the gauge identified with `key` and `subname` with `value`

    :keyword value: The value to set the gauge at, default is 1
    :keyword subname: The subname to report the data to (appended to the
        client name)
    """

    name = key
    if subname:
        name += f".{subname}"

    # We never use the relative changes behaviour so attempt to always make it do the
    # set value behaviour instead.
    if value < 0:
        get_statsd_client().gauge(name, 0)
    return get_statsd_client().gauge(name, value)


def function_long_name(func, extra=None):
    if extra:
        return ".".join([func.__module__, func.__name__, extra])
    else:
        return ".".join([func.__module__, func.__name__])


def auto_named_statsd_timer(function_to_decorate):
    call_name = function_long_name(function_to_decorate, "call")

    @wraps(function_to_decorate)
    def incr_and_call(*args, **kwargs):
        get_statsd_client().incr(call_name)
        return function_to_decorate(*args, **kwargs)

    timer_name = function_long_name(function_to_decorate, "time")
    named_decorator = get_statsd_client().timer(timer_name)

    return named_decorator(incr_and_call)


@contextmanager
def view_metrics_context(endpoint_name=None, instance=None):
    """
    Context manager for view methods to measure hit count and response time.

    Args:
        endpoint_name: Optional custom name for the endpoint. If not provided,
                       uses the instance's class name and current context.
        instance: The instance (self) to get class name from.

    Example:
        async def get(self, request):
            with view_metrics_context(instance=self):
                # Your code here
                return Response({"data": result})

        with view_metrics_context("custom_endpoint"):
            # Your code here
            pass
    """
    if endpoint_name:
        metric_base = f"views.{endpoint_name}"
    else:
        if instance:
            cls_name = instance.__class__.__name__
            metric_base = f"views.{cls_name}.context"
        else:
            metric_base = "views.unknown.context"

    get_statsd_client().incr(f"{metric_base}.hits")

    with get_statsd_client().timer(f"{metric_base}.response_time"):
        yield


def view_metrics(endpoint_name=None):
    """
    Decorator for view methods to measure hit count and response time.

    Args:
        endpoint_name: Optional custom name for the endpoint. If not provided,
                       uses the class name and method name.

    Example:
        @view_metrics()
        async def get(self, request):
            ...

        @view_metrics("custom_endpoint")
        async def post(self, request):
            ...
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            if endpoint_name:
                metric_base = f"views.{endpoint_name}"
            else:
                cls_name = args[0].__class__.__name__ if args else "unknown"
                method_name = func.__name__
                metric_base = f"views.{cls_name}.{method_name}"

            get_statsd_client().incr(f"{metric_base}.hits")

            # Time the function execution
            with get_statsd_client().timer(f"{metric_base}.response_time"):
                return await func(*args, **kwargs)

        return wrapper

    return decorator



def generate_kpi_card_changes(column_name, usd_change=False, percentage_change=False):
    """
    Generate SQL template for calculating KPI changes and percentage changes.

    Generates SQL snippets that calculate:
    - Current value
    - Absolute change (current - previous)
    - Percentage change (if percentage_change is True)
    - USD changes (if usd_change is True), also generates the same calculations for USD
      values.

    Args:
        column_name (str): Name of the column to generate changes for
        usd_change (bool, optional): Whether to include USD calculations. Defaults to
                                     False.
        percentage_change (bool, optional): Whether to include percentage change
                                            calculations. Defaults to False.

    Returns:
        str: SQL template string with calculations for changes and percentage changes

    Example:
        >>> generate_kpi_card_changes('total_supply')
        # Returns SQL template for total_supply and total_supply_usd changes
    """

    template = f"""
        , c.{column_name}
        , c.{column_name} - COALESCE(p.{column_name}, 0) AS {column_name}_change
    """

    if percentage_change:
        template += f"""
            , (
                (c.{column_name} - COALESCE(p.{column_name}, 0))
                / COALESCE(NULLIF(p.{column_name}, 0), NULLIF(c.{column_name}, 0))
            ) AS {column_name}_change_percentage
        """

    if usd_change:
        template += f"""
            , c.{column_name}_usd
            , c.{column_name}_usd - COALESCE(p.{column_name}_usd, 0)
                AS {column_name}_usd_change
        """
        if percentage_change:
            template += f"""
                , (
                    (c.{column_name}_usd - COALESCE(p.{column_name}_usd, 0))
                    / COALESCE(NULLIF(p.{column_name}_usd, 0), NULLIF(
                        c.{column_name}_usd, 0))
                ) AS {column_name}_usd_change_percentage
            """

    return template



# ruff: noqa: T100
import os

from IPython.core.async_helpers import get_asyncio_loop
from IPython.terminal.embed import InteractiveShellEmbed
from tortoise import Tortoise

from bakit import settings


def start_ipython_shell(extra_ns=None, banner=None):
    """
    Async shell helper:
    - starts IPython if available (with top-level await support)
    - falls back to stdlib interactive shell otherwise
    - always closes DB connections when done
    """

    if banner is None:
        banner = "Tortoise shell. If IPython is installed, top-level await should work."

    loop = get_asyncio_loop()

    ns = {
        "Tortoise": Tortoise,
        "os": os,
    }
    if extra_ns:
        ns.update(extra_ns)

    try:
        loop.run_until_complete(
            Tortoise.init(config=settings.TORTOISE_ORM, _enable_global_fallback=True)
        )
        shell = InteractiveShellEmbed(banner2=banner)
        shell(local_ns=ns, global_ns=ns)
    finally:
        loop.run_until_complete(Tortoise.close_connections())
